REBOL [
    Title: "Tic Tac Toe"
    Date: 27-Dec-2001/11:50:39-8:00
    Version: 1.0.3
    File: %tictactoe.r
    Author: "Ryan S. Cole"
    Purpose: "No known purpose."
    History: [
    23-Oct-2001 "Made things more colorful." 
    27-Dec-2001 "Fixed bug allowing win condition." 
    27-Dec-2001/11:50:39-8:00 "Now works properly with View or Link."
    30-Apr-2004 "compatibilit view 1.2.41"
]
    Email: ryancole@usa.com
    Category: [game]
    Comments: {Special thanks to Larry Snyder for help in simplifying game logic.}
]

appli: "REBOL - Tic Tac Toe"
do %uneinstance.r

;;;;;;;;;;;;;;;;;
;;; game data ;;;
;;;;;;;;;;;;;;;;;

; represents the 9 positions available
game-grid: copy "         " ; 9 spaces

; the winning mark combinations
win-table: [
    "mmm??????" "???mmm???" "??????mmm" "m??m??m??"
    "?m??m??m?" "??m??m??m" "m???m???m" "??m?m?m??"
]

;;;;;;;;;;;;;;;;;;;;;;;;
;;; helper functions ;;;
;;;;;;;;;;;;;;;;;;;;;;;;

; returns opposing players marker
other: func [m [char!]][either m = #"X" [#"O"][#"X"]]

; tests for a winning condition for a particular marker
win?: function [grid mark] [tmp-grid] [
    replace/all tmp-grid: copy to-string grid mark #"m"
    foreach win win-table [
        if found? find/any tmp-grid win [return yes]
    ]
]

;;;;;;;;;;;;;;;;;;
;;; game logic ;;;
;;;;;;;;;;;;;;;;;;

; returns the best move
get-win-move: func [grid mark /move] [
	"fonction modifie pour view 1.2.41"
	wait 0.5
    if grid/5 = #" " [poke grid 5 mark return grid]
    if move: win-move? grid mark [poke grid move mark return grid]
    if move: win-move? grid other mark [poke grid move mark return grid]
    if (grid/5 = mark) and any [
        (grid/1 = grid/9) and (grid/1 = other mark)
        (grid/3 = grid/7) and (grid/3 = other mark)
    ] [ 
        poke grid 4 mark return grid
    ]
    either (grid/6 = other mark) or (grid/8 = other mark) [
        if grid/9 = #" " [poke grid 9 mark return grid]
        if grid/7 = #" " [poke grid 7 mark return grid]
        if grid/3 = #" " [poke grid 3 mark return grid]
        if grid/1 = #" " [poke grid 1 mark return grid]
    ] [
        if grid/1 = #" " [poke grid 1 mark return grid]
        if grid/3 = #" " [poke grid 3 mark return grid]
        if grid/7 = #" " [poke grid 7 mark return grid]
        if grid/9 = #" " [poke grid 9 mark return grid]
    ]
    replace grid #" " mark
    return grid
]

; tries to return the move that would result an immediate win for a given marker
win-move?: func [grid mark /local grid-tmp] [
	"fonction modifie pour view 1.2.41"
    repeat idx 9 [
        if grid/:idx = #" " [
        	grid-tmp: copy grid
        	poke grid-tmp idx mark
        	if win? grid-tmp mark [return idx]
            ;if win? poke copy grid idx mark mark [return idx]
        ]
    ]
]

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; game control and visual ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; creates a layout for the screen
lay-grid: does [
    lay: copy [
        backdrop white effect [merge gradcol 1x1 255.0.0 0.0.255]
        origin 0x0  space 3x3
        style square box 100x100 ivory font [color: red size: 40 shadow: none]
    ]
    repeat idx length? game-grid [
        repend lay [ 'square to-string game-grid/:idx ]
        if game-grid/:idx = #"O" [ append lay [font [color: blue] ] ]
        if game-grid/:idx = #" " [ repend/only lay ['react-to-move idx] ]
        if zero? idx // 3 [append lay 'return]
    ]
    append lay [at 0x0 message: label 0x0 center font-size 40]
]

; reacts to users input when a space is pressed
react-to-move: func [index] [
    poke game-grid index #"X"
    check-state
    game-grid: get-win-move game-grid #"O"
    check-state
]

; checks for a lose or tie and updates screen
check-state: does [
    if win? game-grid #"O" [clear-game "PERDU!"]
    if not found? find to-string game-grid #" " [clear-game "EGALITE!"]
    refresh
]

; Clears the game board with special effects.
Clear-game: function [msg] [color] [
    refresh
    wait 1.0
    message/text: msg
    message/size: 300x300
    message/effect: copy []
    color: 255.0.0
    loop 16 [
        message/font/color: color: color - 16.0.0 + 0.0.16
        append message/effect [merge blur]
        show main-face
        wait .2
    ]
    game-grid: copy "         "
]

; repaints the screen
refresh: does [
    main-face/pane: layout/offset lay-grid 0x0
    show main-face
]

;;;;;;;;;;;;;;;;;;;;;;;
;;; start game play ;;;
;;;;;;;;;;;;;;;;;;;;;;;

main-face: view/new center-face layout lay-grid
do-events
                                                                                                                                                  